/* $VER: pasm elfrelnames.c V1.3c (08.02.00)
 *
 * This file is part of pasm, a portable PowerPC assembler.
 * Copyright (c) 1997-2003  Frank Wille
 *
 * pasm is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2003 by Volker Barthelmann.
 * pasm may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 *
 * v1.3c  (08.02.00) phx
 *        New functions get_elfrel_name() for EABI support and
 *        elf_relocsize() which returns the size in bytes for any
 *        relocation type.
 * v1.1   (19.06.98) phx
 *        Deleted R_PPC_TOC16. I think, R_PPC_SDAREL16 is the correct one.
 * v0.4   (02.07.97) phx
 *        Changed name from elfrelocnames.c to elfrelnames.c - SCO 3.2v4
 *        doesn't support more than 15 characters? ;)
 *        File created.
 */


#define ELFRELNAMES_C
#include "ppcasm.h"


static char *elfrel_name[] = {
  "R_NONE",
  "R_PPC_ADDR32",
  "R_PPC_ADDR24",
  "R_PPC_ADDR16",
  "R_PPC_ADDR16_LO",
  "R_PPC_ADDR16_HI",
  "R_PPC_ADDR16_HA",
  "R_PPC_ADDR14",
  "R_PPC_ADDR14_BRTAKEN",
  "R_PPC_ADDR14_BRNTAKEN",
  "R_PPC_REL24",
  "R_PPC_REL14",
  "R_PPC_REL14_BRTAKEN",
  "R_PPC_REL14_BRNTAKEN",
  "R_PPC_GOT16",
  "R_PPC_GOT16_LO",
  "R_PPC_GOT16_HI",
  "R_PPC_GOT16_HA",
  "R_PPC_PLTREL24",
  "R_PPC_COPY",
  "R_PPC_GLOB_DAT",
  "R_PPC_JMP_SLOT",
  "R_PPC_RELATIVE",
  "R_PPC_LOCAL24PC",
  "R_PPC_UADDR32",
  "R_PPC_UADDR16",
  "R_PPC_REL32",
  "R_PPC_PLT32",
  "R_PPC_PLTREL32",
  "R_PPC_PLT16_LO",
  "R_PPC_PLT16_HI",
  "R_PPC_PLT16_HA",
  "R_PPC_SDAREL16",
  "R_PPC_SECTOFF",
  "R_PPC_SECTOFF_LO",
  "R_PPC_SECTOFF_HI",
  "R_PPC_SECTOFF_HA",
  "R_PPC_ADDR30"
};

static char *elfrel_name_eabi[] = {
  NULL,
  "R_PPC_EMB_NADDR32",
  "R_PPC_EMB_NADDR16",
  "R_PPC_EMB_NADDR16_LO",
  "R_PPC_EMB_NADDR16_HI",
  "R_PPC_EMB_NADDR16_HA",
  "R_PPC_EMB_SDAI16",
  "R_PPC_EMB_SDA2I16",
  "R_PPC_EMB_SDA2REL",
  "R_PPC_EMB_SDA21",
  "R_PPC_EMB_MRKREF",
  "R_PPC_EMB_RELSEC16",
  "R_PPC_EMB_RELST_LO",
  "R_PPC_EMB_RELST_HI",
  "R_PPC_EMB_RELST_HA",
  "R_PPC_EMB_BIT_FLD",
  "R_PPC_EMB_RELSDA"
};


char *get_elfrel_name(uint8 rtype)
{
  if (rtype > 100)  /*EABI?*/
    return (elfrel_name_eabi[(rtype-100) & ELFRELNAMMSK]);
  else
    return (elfrel_name[rtype & ELFRELNAMMSK]);
}


int elf_relocsize(uint8 rtype)
{
  switch (rtype) {
    case R_PPC_ADDR32:
    case R_PPC_ADDR24:
    case R_PPC_REL24:
    case R_PPC_GLOB_DAT:
    case R_PPC_RELATIVE:
    case R_PPC_LOCAL24PC:
    case R_PPC_UADDR32:
    case R_PPC_REL32:
    case R_PPC_PLT32:
    case R_PPC_PLTREL32:
    case R_PPC_ADDR30:
    case R_PPC_EMB_NADDR32:
    case R_PPC_EMB_SDA21:
    case R_PPC_EMB_BIT_FLD:
      return 4;
    default:
      return 2;
  }
}
