/* $VER: pasm instructions.c V1.6c (31.10.03)
 *
 * This file is part of pasm, a portable PowerPC assembler.
 * Copyright (c) 1997-2003  Frank Wille
 *
 * pasm is freeware and part of the portable and retargetable ANSI C
 * compiler vbcc, copyright (c) 1995-2003 by Volker Barthelmann.
 * pasm may be freely redistributed as long as no modifications are
 * made and nothing is charged for it. Non-commercial usage is allowed
 * without any restrictions.
 * EVERY PRODUCT OR PROGRAM DERIVED DIRECTLY FROM MY SOURCE MAY NOT BE
 * SOLD COMMERCIALLY WITHOUT PERMISSION FROM THE AUTHOR.
 *
 *
 * v1.6c  (31.10.03) phx
 *        Rotate instructions with only four arguments will regard
 *        the last argument as mask (gas-compatibility).
 * v1.6b  (04.07.03) phx
 *        Code cleanup: removed forward references.
 * v1.6   (23.08.01) phx
 *        AltiVec instruction set.
 * v1.5   (26.01.01) phx
 *        Common symbols are treated as external references.
 * v1.4   (26.11.00) phx
 *        Fixed store_half/word/float/double(), which previously crashed
 *        when writing unaligned data to memory.
 *        Fixed branch prediction. y-bit has to be 0 by default!
 * v1.3a  (31.01.00) phx
 *        mftb without tb register number defaults to 268.
 * v1.3   (28.12.99) phx
 *        Fixed a lot of size-warnings for instructions using "@l". All
 *        T_DD- and T_DS-type instructions have got the F_SIGNED flag.
 * v1.2b  (01.02.99) phx
 *        Fixed move_symbols() for move-deltas > alignment (e.g. .align 2)
 * v1.2   (21.10.98) phx
 *        @sdarx support for T_DD- and T_DS-type instructions.
 *        Replaced obsolete R_PPC_TOC16 by real R_PPC_SDAREL16.
 *        Fixed mfdcr and mtdcr.
 * v1.1d  (27.09.98) phx
 *        Optional PPC4xx instructions mfdcr and mtdcr.
 * v1.1a  (20.06.98) phx
 *        F_UPDATE. Check rA = rD for update instructions.
 * v1.1   (19.06.98) phx
 *        Multiple small data sections are supported. The new directive
 *        .sdreg should be used to set the base register.
 *        Fixed "Extra characters on line" error for "la rN,<symbol>".
 *        Wrong error message (46) for "la rN,<absSymbol>". Changed to
 *        error 32: reloc symbol required.
 * v1.0   (21.05.98) phx
 *        tlsync should have been tlbsync.
 *        mtsrin is a supervisor instruction.
 *        ME was not checked for rlwxxx.
 *        mb/me in rldicxx is 6 bit, not 5.
 * v0.9   (07.03.98) phx
 *        Base relative access on external symbol generated wrong
 *        relocation type.
 *        R_ADDR16 relocations in code section had wrong offset.
 *        R_REL14 branches will only get a 2-byte offset, if the output
 *        format is EHF/ADOS and a relocation- or xref-entry is required.
 * v0.8   (14.02.98) phx
 *        Alignment list for each section. This fixes the problems
 *        with optimizations.
 * v0.7   (02.01.98) phx
 *        Implemented optimization OPT_FAR_BRANCH.
 * v0.6   (30.10.97) phx
 *        No more warnings for val@l(rA), where val is > 0x7fff.
 *        Warnings for optional, 64-bit and supervisor instructions
 *        can be suppressed.
 * v0.5   (03.10.97) phx
 *        Auto-converting (rA) into 0(rA) had lead to conflicts with terms.
 *        For example: (x+y)(rA)
 * v0.4   (05.07.97) phx
 *        AA bit in BC instructions is set to allow correct EHF reloction
 *        (by using AmigaDOS HUNK_RELRELOC16). This should be no problem,
 *        because absolute branches are not supported by EHF anyway.
 *        Branch prediction with BCA didn't work correctly.
 *        Relative branches into other sections are allowed, but the
 *        user will be warned, because not all object formats / linkers
 *        support it.
 *        Omitting zero in load/store instructions is allowed, e.g.
 *        "rD,(rA)" will be converted into "rD,0(rA)".
 *        R_PPC_TOC16 support for T_DD- and T_DS-type instructions.
 * v0.3   (05.04.97) phx
 *        An external symbol@l/h/ha, used in d(Rn) addressing mode, got
 *        R_PPC_ADDR16 relocation instead R_PPC_ADDR16_LO/HI/HA.
 *        Little-endian support.
 * v0.2   (25.03.97) phx
 *        Writes ELF object for 32-bit PowerPC big-endian. Either absolute
 *        or ELF output format may be selected. ELF is default for all
 *        currently supported platforms. PPCasm supports nine different
 *        relocation types (there are much more...).
 *        Compiles and works also under NetBSD/amiga (68k).
 *        Changed function declaration to 'new style' in all sources
 *        (to avoid problems with '...' for example).
 *        @l/h/ha is allowed for all displacements.
 * v0.1   (11.03.97) phx
 *        First test version with all PowerPC instructions and most
 *        important directives. Only raw, absolute output.
 *        Although all 32- and 64-bit PowerPC instructions are imple-
 *        mented, too few are really tested till now. I'm expecting many
 *        bugs in this long list.
 *        'la' is the only extended mnemonic, which was implemented as a
 *        real instruction and not by a macro. It is possible to imple-
 *        ment all extended mnemonics here, but it would be much work.
 *        Maybe later, to improve speed. ;)
 * v0.0   (21.02.97) phx
 *        File created.
 */


#define INSTRUCTIONS_C
#include "ppcasm.h"


struct CPUInstr instructions[] = {
  /* standard PowerPC instruction set */
  { 0,"add",0,T_X,31,0,0,(0<<10)+(266<<1)+0 },
  { 0,"add.",0,T_X,31,0,0,(0<<10)+(266<<1)+1 },
  { 0,"addo",0,T_X,31,0,0,(1<<10)+(266<<1)+0 },
  { 0,"addo.",0,T_X,31,0,0,(1<<10)+(266<<1)+1 },
  { 0,"addc",0,T_X,31,0,0,(0<<10)+(10<<1)+0 },
  { 0,"addc.",0,T_X,31,0,0,(0<<10)+(10<<1)+1 },
  { 0,"addco",0,T_X,31,0,0,(1<<10)+(10<<1)+0 },
  { 0,"addco.",0,T_X,31,0,0,(1<<10)+(10<<1)+1 },
  { 0,"adde",0,T_X,31,0,0,(0<<10)+(138<<1)+0 },
  { 0,"adde.",0,T_X,31,0,0,(0<<10)+(138<<1)+1 },
  { 0,"addeo",0,T_X,31,0,0,(1<<10)+(138<<1)+0 },
  { 0,"addeo.",0,T_X,31,0,0,(1<<10)+(138<<1)+1 },
  { 0,"addi",F_SIGNED,T_DI,14,0,0,0 },
  { 0,"addic",F_SIGNED,T_DI,12,0,0,0 },
  { 0,"addic.",F_SIGNED,T_DI,13,0,0,0 },
  { 0,"addis",F_SIGNED,T_DI,15,0,0,0 },
  { 0,"addme",F_SUPP_B,T_X,31,0,0,(0<<10)+(234<<1)+0 },
  { 0,"addme.",F_SUPP_B,T_X,31,0,0,(0<<10)+(234<<1)+1 },
  { 0,"addmeo",F_SUPP_B,T_X,31,0,0,(1<<10)+(234<<1)+0 },
  { 0,"addmeo.",F_SUPP_B,T_X,31,0,0,(1<<10)+(234<<1)+1 },
  { 0,"addze",F_SUPP_B,T_X,31,0,0,(0<<10)+(202<<1)+0 },
  { 0,"addze.",F_SUPP_B,T_X,31,0,0,(0<<10)+(202<<1)+1 },
  { 0,"addzeo",F_SUPP_B,T_X,31,0,0,(1<<10)+(202<<1)+0 },
  { 0,"addzeo.",F_SUPP_B,T_X,31,0,0,(1<<10)+(202<<1)+1 },
  { 0,"and",F_SWAP,T_X,31,0,0,(28<<1)+0 },
  { 0,"and.",F_SWAP,T_X,31,0,0,(28<<1)+1 },
  { 0,"andc",F_SWAP,T_X,31,0,0,(60<<1)+0 },
  { 0,"andc.",F_SWAP,T_X,31,0,0,(60<<1)+1 },
  { 0,"andi.",F_SWAP,T_DI,28,0,0,0 },
  { 0,"andis.",F_SWAP,T_DI,29,0,0,0 },
  { 0,"b",F_SIGNED,T_I,18,0,0,0 },
  { 0,"bl",F_SIGNED,T_I,18,0,0,1 },
  { 0,"ba",0,T_I,18,0,0,2 },
  { 0,"bla",0,T_I,18,0,0,3 },
  { 0,"bc",F_SIGNED,T_B,16,0,0,0 },
  { 0,"bcl",F_SIGNED,T_B,16,0,0,1 },
  { 0,"bca",0,T_B,16,0,0,2 },
  { 0,"bcla",0,T_B,16,0,0,3 },
  { 0,"bcctr",0,T_XLB,19,0,0,(528<<1)+0 },
  { 0,"bcctrl",0,T_XLB,19,0,0,(528<<1)+1 },
  { 0,"bclr",0,T_XLB,19,0,0,(16<<1)+0 },
  { 0,"bclrl",0,T_XLB,19,0,0,(16<<1)+1 },
  { 0,"cmp",0,T_CMP,31,0,0,(0<<1) },
  { 0,"cmpl",0,T_CMP,31,0,0,(32<<1) },
  { 0,"cmpi",F_SIGNED|F_SUPP_B,T_CMP,11,0,0,0 },
  { 0,"cmpli",F_SUPP_B,T_CMP,10,0,0,0 },
  { 0,"cntlzd",F_SWAP|F_SUPP_B|F_64BIT,T_X,31,0,0,(58<<1)+0 },
  { 0,"cntlzd.",F_SWAP|F_SUPP_B|F_64BIT,T_X,31,0,0,(58<<1)+1 },
  { 0,"cntlzw",F_SWAP|F_SUPP_B,T_X,31,0,0,(26<<1)+0 },
  { 0,"cntlzw.",F_SWAP|F_SUPP_B,T_X,31,0,0,(26<<1)+1 },
  { 0,"crand",0,T_X,19,0,0,(257<<1) },
  { 0,"crandc",0,T_X,19,0,0,(129<<1) },
  { 0,"creqv",0,T_X,19,0,0,(289<<1) },
  { 0,"crnand",0,T_X,19,0,0,(225<<1) },
  { 0,"crnor",0,T_X,19,0,0,(33<<1) },
  { 0,"cror",0,T_X,19,0,0,(449<<1) },
  { 0,"crorc",0,T_X,19,0,0,(417<<1) },
  { 0,"crxor",0,T_X,19,0,0,(193<<1) },
  { 0,"dcbf",F_SUPP_D,T_X,31,0,0,(86<<1) },
  { 0,"dcbi",F_SUPP_D|F_SUPER,T_X,31,0,0,(470<<1) },
  { 0,"dcbst",F_SUPP_D,T_X,31,0,0,(54<<1) },
  { 0,"dcbt",F_SUPP_D,T_X,31,0,0,(278<<1) },
  { 0,"dcbtst",F_SUPP_D,T_X,31,0,0,(246<<1) },
  { 0,"dcbz",F_SUPP_D,T_X,31,0,0,(1014<<1) },
  { 0,"divd",F_64BIT,T_X,31,0,0,(0<<10)+(489<<1)+0 },
  { 0,"divd.",F_64BIT,T_X,31,0,0,(0<<10)+(489<<1)+1 },
  { 0,"divdo",F_64BIT,T_X,31,0,0,(1<<10)+(489<<1)+0 },
  { 0,"divdo.",F_64BIT,T_X,31,0,0,(1<<10)+(489<<1)+1 },
  { 0,"divdu",F_64BIT,T_X,31,0,0,(0<<10)+(457<<1)+0 },
  { 0,"divdu.",F_64BIT,T_X,31,0,0,(0<<10)+(457<<1)+1 },
  { 0,"divduo",F_64BIT,T_X,31,0,0,(1<<10)+(457<<1)+0 },
  { 0,"divduo.",F_64BIT,T_X,31,0,0,(1<<10)+(457<<1)+1 },
  { 0,"divw",0,T_X,31,0,0,(0<<10)+(491<<1)+0 },
  { 0,"divw.",0,T_X,31,0,0,(0<<10)+(491<<1)+1 },
  { 0,"divwo",0,T_X,31,0,0,(1<<10)+(491<<1)+0 },
  { 0,"divwo.",0,T_X,31,0,0,(1<<10)+(491<<1)+1 },
  { 0,"divwu",0,T_X,31,0,0,(0<<10)+(459<<1)+0 },
  { 0,"divwu.",0,T_X,31,0,0,(0<<10)+(459<<1)+1 },
  { 0,"divwuo",0,T_X,31,0,0,(1<<10)+(459<<1)+0 },
  { 0,"divwuo.",0,T_X,31,0,0,(1<<10)+(459<<1)+1 },
  { 0,"eciwx",0,T_X,31,0,0,(310<<1) },
  { 0,"ecowx",0,T_X,31,0,0,(438<<1) },
  { 0,"eieio",F_SUPP_D|F_SUPP_A|F_SUPP_B,T_X,31,0,0,(854<<1) },
  { 0,"eqv",F_SWAP,T_X,31,0,0,(284<<1)+0 },
  { 0,"eqv.",F_SWAP,T_X,31,0,0,(284<<1)+0 },
  { 0,"extsb",F_SWAP|F_SUPP_B,T_X,31,0,0,(954<<1)+0 },
  { 0,"extsb.",F_SWAP|F_SUPP_B,T_X,31,0,0,(954<<1)+1 },
  { 0,"extsh",F_SWAP|F_SUPP_B,T_X,31,0,0,(922<<1)+0 },
  { 0,"extsh.",F_SWAP|F_SUPP_B,T_X,31,0,0,(922<<1)+1 },
  { 0,"extsw",F_SWAP|F_SUPP_B|F_64BIT,T_X,31,0,0,(986<<1)+0 },
  { 0,"extsw.",F_SWAP|F_SUPP_B|F_64BIT,T_X,31,0,0,(986<<1)+1 },
  { 0,"fabs",F_SUPP_A,T_X,63,0,0,(264<<1)+0 },
  { 0,"fabs.",F_SUPP_A,T_X,63,0,0,(264<<1)+1 },
  { 0,"fadd",F_SUPP_C,T_A,63,0,0,(21<<1)+0 },
  { 0,"fadd.",F_SUPP_C,T_A,63,0,0,(21<<1)+1 },
  { 0,"fadds",F_SUPP_C,T_A,59,0,0,(21<<1)+0 },
  { 0,"fadds.",F_SUPP_C,T_A,59,0,0,(21<<1)+1 },
  { 0,"fcfid",F_SUPP_A|F_64BIT,T_X,63,0,0,(846<<1)+0 },
  { 0,"fcfid.",F_SUPP_A|F_64BIT,T_X,63,0,0,(846<<1)+1 },
  { 0,"fcmpo",F_CRF_D,T_X,63,0,0,(32<<1) },
  { 0,"fcmpu",F_CRF_D,T_X,63,0,0,(0<<1) },
  { 0,"fctid",F_SUPP_A|F_64BIT,T_X,63,0,0,(814<<1)+0 },
  { 0,"fctid.",F_SUPP_A|F_64BIT,T_X,63,0,0,(814<<1)+1 },
  { 0,"fctidz",F_SUPP_A|F_64BIT,T_X,63,0,0,(815<<1)+0 },
  { 0,"fctidz.",F_SUPP_A|F_64BIT,T_X,63,0,0,(815<<1)+1 },
  { 0,"fctiw",F_SUPP_A,T_X,63,0,0,(14<<1)+0 },
  { 0,"fctiw.",F_SUPP_A,T_X,63,0,0,(14<<1)+1 },
  { 0,"fctiwz",F_SUPP_A,T_X,63,0,0,(15<<1)+0 },
  { 0,"fctiwz.",F_SUPP_A,T_X,63,0,0,(15<<1)+1 },
  { 0,"fdiv",F_SUPP_C,T_A,63,0,0,(18<<1)+0 },
  { 0,"fdiv.",F_SUPP_C,T_A,63,0,0,(18<<1)+1 },
  { 0,"fdivs",F_SUPP_C,T_A,59,0,0,(18<<1)+0 },
  { 0,"fdivs.",F_SUPP_C,T_A,59,0,0,(18<<1)+1 },
  { 0,"fmadd",0,T_A,63,0,0,(29<<1)+0 },
  { 0,"fmadd.",0,T_A,63,0,0,(29<<1)+1 },
  { 0,"fmadds",0,T_A,59,0,0,(29<<1)+0 },
  { 0,"fmadds.",0,T_A,59,0,0,(29<<1)+1 },
  { 0,"fmr",F_SUPP_A,T_X,63,0,0,(72<<1)+0 },
  { 0,"fmr.",F_SUPP_A,T_X,63,0,0,(72<<1)+1 },
  { 0,"fmsub",0,T_A,63,0,0,(28<<1)+0 },
  { 0,"fmsub.",0,T_A,63,0,0,(28<<1)+1 },
  { 0,"fmsubs",0,T_A,59,0,0,(28<<1)+0 },
  { 0,"fmsubs.",0,T_A,59,0,0,(28<<1)+1 },
  { 0,"fmul",F_SUPP_B,T_A,63,0,0,(25<<1)+0 },
  { 0,"fmul.",F_SUPP_B,T_A,63,0,0,(25<<1)+1 },
  { 0,"fmuls",F_SUPP_B,T_A,59,0,0,(25<<1)+0 },
  { 0,"fmuls.",F_SUPP_B,T_A,59,0,0,(25<<1)+1 },
  { 0,"fnabs",F_SUPP_A,T_X,63,0,0,(136<<1)+0 },
  { 0,"fnabs.",F_SUPP_A,T_X,63,0,0,(136<<1)+1 },
  { 0,"fneg",F_SUPP_A,T_X,63,0,0,(40<<1)+0 },
  { 0,"fneg.",F_SUPP_A,T_X,63,0,0,(40<<1)+1 },
  { 0,"fnmadd",0,T_A,63,0,0,(31<<1)+0 },
  { 0,"fnmadd.",0,T_A,63,0,0,(31<<1)+1 },
  { 0,"fnmadds",0,T_A,59,0,0,(31<<1)+0 },
  { 0,"fnmadds.",0,T_A,59,0,0,(31<<1)+1 },
  { 0,"fnmsub",0,T_A,63,0,0,(30<<1)+0 },
  { 0,"fnmsub.",0,T_A,63,0,0,(30<<1)+1 },
  { 0,"fnmsubs",0,T_A,59,0,0,(30<<1)+0 },
  { 0,"fnmsubs.",0,T_A,59,0,0,(30<<1)+1 },
  { 0,"fres",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,59,0,0,(24<<1)+0 },
  { 0,"fres.",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,59,0,0,(24<<1)+1 },
  { 0,"frsp",F_SUPP_A,T_X,63,0,0,(12<<1)+0 },
  { 0,"frsp.",F_SUPP_A,T_X,63,0,0,(12<<1)+1 },
  { 0,"frsqrte",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,63,0,0,(26<<1)+0 },
  { 0,"frsqrte.",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,63,0,0,(26<<1)+1 },
  { 0,"fsel",F_OPTIONAL,T_A,63,0,0,(23<<1)+0 },
  { 0,"fsel.",F_OPTIONAL,T_A,63,0,0,(23<<1)+1 },
  { 0,"fsqrt",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,63,0,0,(22<<1)+0 },
  { 0,"fsqrt.",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,63,0,0,(22<<1)+1 },
  { 0,"fsqrts",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,59,0,0,(22<<1)+0 },
  { 0,"fsqrts.",F_SUPP_A|F_SUPP_C|F_OPTIONAL,T_A,59,0,0,(22<<1)+1 },
  { 0,"fsub",F_SUPP_C,T_A,63,0,0,(20<<1)+0 },
  { 0,"fsub.",F_SUPP_C,T_A,63,0,0,(20<<1)+1 },
  { 0,"fsubs",F_SUPP_C,T_A,59,0,0,(20<<1)+0 },
  { 0,"fsubs.",F_SUPP_C,T_A,59,0,0,(20<<1)+1 },
  { 0,"icbi",F_SUPP_D,T_X,31,0,0,(982<<1) },
  { 0,"isync",F_SUPP_D|F_SUPP_A|F_SUPP_B,T_X,19,0,0,(150<<1) },
  { 0,"lbz",F_SIGNED,T_DD,34,0,0,0 },
  { 0,"lbzu",F_SIGNED|F_UPDATE,T_DD,35,0,0,0 },
  { 0,"lbzux",F_UPDATE,T_X,31,0,0,(119<<1) },
  { 0,"lbzx",0,T_X,31,0,0,(87<<1) },
  { 0,"ld",F_SIGNED|F_64BIT,T_DS,58,0,0,0 },
  { 0,"ldarx",F_64BIT,T_X,31,0,0,(84<<1) },
  { 0,"ldu",F_SIGNED|F_64BIT|F_UPDATE,T_DS,58,0,0,1 },
  { 0,"ldux",F_64BIT|F_UPDATE,T_X,31,0,0,(53<<1) },
  { 0,"ldx",F_64BIT,T_X,31,0,0,(21<<1) },
  { 0,"lfd",F_SIGNED,T_DD,50,0,0,0 },
  { 0,"lfdu",F_SIGNED,T_DD,51,0,0,0 },
  { 0,"lfdux",0,T_X,31,0,0,(631<<1) },
  { 0,"lfdx",0,T_X,31,0,0,(599<<1) },
  { 0,"lfs",F_SIGNED,T_DD,48,0,0,0 },
  { 0,"lfsu",F_SIGNED,T_DD,49,0,0,0 },
  { 0,"lfsux",0,T_X,31,0,0,(567<<1) },
  { 0,"lfsx",0,T_X,31,0,0,(535<<1) },
  { 0,"lha",F_SIGNED,T_DD,42,0,0,0 },
  { 0,"lhau",F_SIGNED|F_UPDATE,T_DD,43,0,0,0 },
  { 0,"lhaux",F_UPDATE,T_X,31,0,0,(375<<1) },
  { 0,"lhax",0,T_X,31,0,0,(343<<1) },
  { 0,"lhbrx",0,T_X,31,0,0,(790<<1) },
  { 0,"lhz",F_SIGNED,T_DD,40,0,0,0 },
  { 0,"lhzu",F_SIGNED|F_UPDATE,T_DD,41,0,0,0 },
  { 0,"lhzux",F_UPDATE,T_X,31,0,0,(311<<1) },
  { 0,"lhzx",0,T_X,31,0,0,(279<<1) },
  { 0,"lmw",F_SIGNED,T_DD,46,0,0,0 },
  { 0,"lswi",0,T_X,31,0,0,(597<<1) },
  { 0,"lswx",0,T_X,31,0,0,(533<<1) },
  { 0,"lwa",F_SIGNED|F_64BIT,T_DS,58,0,0,2 },
  { 0,"lwarx",0,T_X,31,0,0,(20<<1) },
  { 0,"lwaux",F_64BIT|F_UPDATE,T_X,31,0,0,(373<<1) },
  { 0,"lwax",F_64BIT,T_X,31,0,0,(341<<1) },
  { 0,"lwbrx",0,T_X,31,0,0,(534<<1) },
  { 0,"lwz",F_SIGNED,T_DD,32,0,0,0 },
  { 0,"lwzu",F_SIGNED|F_UPDATE,T_DD,33,0,0,0 },
  { 0,"lwzux",F_UPDATE,T_X,31,0,0,(55<<1) },
  { 0,"lwzx",0,T_X,31,0,0,(23<<1) },
  { 0,"mcrf",F_CRF_D|F_CRF_S|F_SUPP_B,T_X,19,0,0,(0<<1) },
  { 0,"mcrfs",F_CRF_D|F_CRF_S|F_SUPP_B,T_X,63,0,0,(64<<1) },
  { 0,"mcrxr",F_CRF_D|F_SUPP_A|F_SUPP_B,T_X,31,0,0,(512<<1) },
  { 0,"mfcr",F_SUPP_A|F_SUPP_B,T_X,31,0,0,(19<<1) },
  { 0,"mffs",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(583<<1)+0 },
  { 0,"mffs.",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(583<<1)+1 },
  { 0,"mfmsr",F_SUPP_A|F_SUPP_B|F_SUPER,T_X,31,0,0,(83<<1) },
  { 0,"mfdcr",F_OPTIONAL,T_XSPR,31,0,0,(323<<1) },  /* PPC4xx - jab */
  { 0,"mfspr",0,T_XSPR,31,0,0,(339<<1) },
  { 0,"mfsr",F_SUPP_B|F_SUPER,T_X,31,0,0,(595<<1) },
  { 0,"mfsrin",F_SUPP_A|F_SUPER,T_X,31,0,0,(659<<1) },
  { 0,"mftb",0,T_XSPR,31,0,0,(371<<1) },
  { 0,"mtcrf",0,T_XCRM,31,0,0,(144<<1) },
  { 0,"mtfsb0",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(70<<1)+0 },
  { 0,"mtfsb0.",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(70<<1)+1 },
  { 0,"mtfsb1",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(38<<1)+0 },
  { 0,"mtfsb1.",F_SUPP_A|F_SUPP_B,T_X,63,0,0,(38<<1)+1 },
  { 0,"mtfsf",0,T_XFL,63,0,0,(711<<1)+0 },
  { 0,"mtfsf.",0,T_XFL,63,0,0,(711<<1)+1 },
  { 0,"mtfsfi",0,T_IMM,63,0,0,(134<<1)+0 },
  { 0,"mtfsfi.",0,T_IMM,63,0,0,(134<<1)+1 },
  { 0,"mtmsr",F_SUPP_A|F_SUPP_B|F_SUPER,T_X,31,0,0,(146<<1) },
  { 0,"mtdcr",F_SWAP|F_OPTIONAL,T_XSPR,31,0,0,(451<<1) },  /* PPC4xx - jab */
  { 0,"mtspr",F_SWAP,T_XSPR,31,0,0,(467<<1) },
  { 0,"mtsr",F_SWAP|F_SUPP_B|F_SUPER,T_X,31,0,0,(210<<1) },
  { 0,"mtsrin",F_SUPP_A|F_SUPER,T_X,31,0,0,(242<<1) },
  { 0,"mulhd",F_64BIT,T_X,31,0,0,(73<<1)+0 },
  { 0,"mulhd.",F_64BIT,T_X,31,0,0,(73<<1)+1 },
  { 0,"mulhdu",F_64BIT,T_X,31,0,0,(9<<1)+0 },
  { 0,"mulhdu.",F_64BIT,T_X,31,0,0,(9<<1)+1 },
  { 0,"mulhw",0,T_X,31,0,0,(75<<1)+0 },
  { 0,"mulhw.",0,T_X,31,0,0,(75<<1)+1 },
  { 0,"mulhwu",0,T_X,31,0,0,(11<<1)+0 },
  { 0,"mulhwu.",0,T_X,31,0,0,(11<<1)+1 },
  { 0,"mulld",F_64BIT,T_X,31,0,0,(0<<10)+(233<<1)+0 },
  { 0,"mulld.",F_64BIT,T_X,31,0,0,(0<<10)+(233<<1)+1 },
  { 0,"mulldo",F_64BIT,T_X,31,0,0,(1<<10)+(233<<1)+0 },
  { 0,"mulldo.",F_64BIT,T_X,31,0,0,(1<<10)+(233<<1)+1 },
  { 0,"mulli",F_SIGNED,T_DI,7,0,0,0 },
  { 0,"mullw",0,T_X,31,0,0,(0<<10)+(235<<1)+0 },
  { 0,"mullw.",0,T_X,31,0,0,(0<<10)+(235<<1)+1 },
  { 0,"mullwo",0,T_X,31,0,0,(1<<10)+(235<<1)+0 },
  { 0,"mullwo.",0,T_X,31,0,0,(1<<10)+(235<<1)+1 },
  { 0,"nand",F_SWAP,T_X,31,0,0,(476<<1)+0 },
  { 0,"nand.",F_SWAP,T_X,31,0,0,(476<<1)+1 },
  { 0,"neg",F_SUPP_B,T_X,31,0,0,(0<<10)+(104<<1)+0 },
  { 0,"neg.",F_SUPP_B,T_X,31,0,0,(0<<10)+(104<<1)+1 },
  { 0,"nego",F_SUPP_B,T_X,31,0,0,(1<<10)+(104<<1)+0 },
  { 0,"nego.",F_SUPP_B,T_X,31,0,0,(1<<10)+(104<<1)+1 },
  { 0,"nor",F_SWAP,T_X,31,0,0,(124<<1)+0 },
  { 0,"nor.",F_SWAP,T_X,31,0,0,(124<<1)+1 },
  { 0,"or",F_SWAP,T_X,31,0,0,(444<<1)+0 },
  { 0,"or.",F_SWAP,T_X,31,0,0,(444<<1)+1 },
  { 0,"orc",F_SWAP,T_X,31,0,0,(412<<1)+0 },
  { 0,"orc.",F_SWAP,T_X,31,0,0,(412<<1)+1 },
  { 0,"ori",F_SWAP,T_DI,24,0,0,0 },
  { 0,"oris",F_SWAP,T_DI,25,0,0,0 },
  { 0,"rfi",F_SUPP_D|F_SUPP_A|F_SUPP_B|F_SUPER,T_X,19,0,0,(50<<1) },
  { 0,"rldcl",F_SWAP|F_64BIT,T_A,30,0,0,(8<<1)+0 },
  { 0,"rldcl.",F_SWAP|F_64BIT,T_A,30,0,0,(8<<1)+1 },
  { 0,"rldcr",F_SWAP|F_64BIT,T_A,30,0,0,(9<<1)+0 },
  { 0,"rldcr.",F_SWAP|F_64BIT,T_A,30,0,0,(9<<1)+1 },
  { 0,"rldic",F_64BIT,T_MD,30,0,0,(2<<2)+0 },
  { 0,"rldic.",F_64BIT,T_MD,30,0,0,(2<<2)+1 },
  { 0,"rldicl",F_64BIT,T_MD,30,0,0,(0<<2)+0 },
  { 0,"rldicl.",F_64BIT,T_MD,30,0,0,(0<<2)+1 },
  { 0,"rldicr",F_64BIT,T_MD,30,0,0,(1<<2)+0 },
  { 0,"rldicr.",F_64BIT,T_MD,30,0,0,(1<<2)+1 },
  { 0,"rldimi",F_64BIT,T_MD,30,0,0,(3<<2)+0 },
  { 0,"rldimi.",F_64BIT,T_MD,30,0,0,(3<<2)+1 },
  { 0,"rlwimi",0,T_M,20,0,0,0 },
  { 0,"rlwimi.",0,T_M,20,0,0,1 },
  { 0,"rlwinm",0,T_M,21,0,0,0 },
  { 0,"rlwinm.",0,T_M,21,0,0,1 },
  { 0,"rlwnm",0,T_M,23,0,0,0 },
  { 0,"rlwnm.",0,T_M,23,0,0,1 },
  { 0,"sc",F_SUPP_D|F_SUPP_A|F_SUPP_B,T_X,17,0,0,2 },
  { 0,"slbia",F_SUPP_D|F_SUPP_A|F_SUPP_B|F_SUPER|F_64BIT|F_OPTIONAL,
    T_X,31,0,0,(498<<1) },
  { 0,"slbie",F_SUPP_D|F_SUPP_A|F_SUPER|F_64BIT|F_OPTIONAL,
    T_X,31,0,0,(434<<1) },
  { 0,"sld",F_SWAP|F_64BIT,T_X,31,0,0,(27<<1)+0 },
  { 0,"sld.",F_SWAP|F_64BIT,T_X,31,0,0,(27<<1)+1 },
  { 0,"slw",F_SWAP,T_X,31,0,0,(24<<1)+0 },
  { 0,"slw.",F_SWAP,T_X,31,0,0,(24<<1)+1 },
  { 0,"srad",F_SWAP|F_64BIT,T_X,31,0,0,(794<<1)+0 },
  { 0,"srad.",F_SWAP|F_64BIT,T_X,31,0,0,(794<<1)+1 },
  { 0,"sradi",F_64BIT,T_XS,31,0,0,(413<<2)+0 },
  { 0,"sradi.",F_64BIT,T_XS,31,0,0,(413<<2)+1 },
  { 0,"sraw",F_SWAP,T_X,31,0,0,(792<<1)+0 },
  { 0,"sraw.",F_SWAP,T_X,31,0,0,(792<<1)+1 },
  { 0,"srawi",F_SWAP,T_X,31,0,0,(824<<1)+0 },
  { 0,"srawi.",F_SWAP,T_X,31,0,0,(824<<1)+1 },
  { 0,"srd",F_SWAP|F_64BIT,T_X,31,0,0,(539<<1)+0 },
  { 0,"srd.",F_SWAP|F_64BIT,T_X,31,0,0,(539<<1)+1 },
  { 0,"srw",F_SWAP,T_X,31,0,0,(536<<1)+0 },
  { 0,"srw.",F_SWAP,T_X,31,0,0,(536<<1)+1 },
  { 0,"stb",F_SIGNED,T_DD,38,0,0,0 },
  { 0,"stbu",F_SIGNED,T_DD,39,0,0,0 },
  { 0,"stbux",0,T_X,31,0,0,(247<<1) },
  { 0,"stbx",0,T_X,31,0,0,(215<<1) },
  { 0,"std",F_SIGNED|F_64BIT,T_DS,62,0,0,0 },
  { 0,"stdcx.",F_64BIT,T_X,31,0,0,(214<<1)+1 },
  { 0,"stdu",F_SIGNED|F_64BIT,T_DS,62,0,0,1 },
  { 0,"stdux",F_64BIT,T_X,31,0,0,(181<<1) },
  { 0,"stdx",F_64BIT,T_X,31,0,0,(149<<1) },
  { 0,"stfd",F_SIGNED,T_DD,54,0,0,0 },
  { 0,"stfdu",F_SIGNED,T_DD,55,0,0,0 },
  { 0,"stfdux",0,T_X,31,0,0,(759<<1) },
  { 0,"stfdx",0,T_X,31,0,0,(727<<1) },
  { 0,"stfiwx",F_OPTIONAL,T_X,31,0,0,(983<<1) },
  { 0,"stfs",F_SIGNED,T_DD,52,0,0,0 },
  { 0,"stfsu",F_SIGNED,T_DD,53,0,0,0 },
  { 0,"stfsux",0,T_X,31,0,0,(695<<1) },
  { 0,"stfsx",0,T_X,31,0,0,(663<<1) },
  { 0,"sth",F_SIGNED,T_DD,44,0,0,0 },
  { 0,"sthbrx",0,T_X,31,0,0,(918<<1) },
  { 0,"sthu",F_SIGNED,T_DD,45,0,0,0 },
  { 0,"sthux",0,T_X,31,0,0,(439<<1) },
  { 0,"sthx",0,T_X,31,0,0,(407<<1) },
  { 0,"stmw",F_SIGNED,T_DD,47,0,0,0 },
  { 0,"stswi",0,T_X,31,0,0,(725<<1) },
  { 0,"stswx",0,T_X,31,0,0,(661<<1) },
  { 0,"stw",F_SIGNED,T_DD,36,0,0,0 },
  { 0,"stwbrx",0,T_X,31,0,0,(662<<1) },
  { 0,"stwcx.",0,T_X,31,0,0,(150<<1)+1 },
  { 0,"stwu",F_SIGNED,T_DD,37,0,0,0 },
  { 0,"stwux",0,T_X,31,0,0,(183<<1) },
  { 0,"stwx",0,T_X,31,0,0,(151<<1) },
  { 0,"subf",0,T_X,31,0,0,(0<<10)+(40<<1)+0 },
  { 0,"subf.",0,T_X,31,0,0,(0<<10)+(40<<1)+1 },
  { 0,"subfo",0,T_X,31,0,0,(1<<10)+(40<<1)+0 },
  { 0,"subfo.",0,T_X,31,0,0,(1<<10)+(40<<1)+1 },
  { 0,"subfc",0,T_X,31,0,0,(0<<10)+(8<<1)+0 },
  { 0,"subfc.",0,T_X,31,0,0,(0<<10)+(8<<1)+1 },
  { 0,"subfco",0,T_X,31,0,0,(1<<10)+(8<<1)+0 },
  { 0,"subfco.",0,T_X,31,0,0,(1<<10)+(8<<1)+1 },
  { 0,"subfe",0,T_X,31,0,0,(0<<10)+(136<<1)+0 },
  { 0,"subfe.",0,T_X,31,0,0,(0<<10)+(136<<1)+1 },
  { 0,"subfeo",0,T_X,31,0,0,(1<<10)+(136<<1)+0 },
  { 0,"subfeo.",0,T_X,31,0,0,(1<<10)+(136<<1)+1 },
  { 0,"subfic",F_SIGNED,T_DI,8,0,0,0 },
  { 0,"subfme",F_SUPP_B,T_X,31,0,0,(0<<10)+(232<<1)+0 },
  { 0,"subfme.",F_SUPP_B,T_X,31,0,0,(0<<10)+(232<<1)+1 },
  { 0,"subfmeo",F_SUPP_B,T_X,31,0,0,(1<<10)+(232<<1)+0 },
  { 0,"subfmeo.",F_SUPP_B,T_X,31,0,0,(1<<10)+(232<<1)+1 },
  { 0,"subfze",F_SUPP_B,T_X,31,0,0,(0<<10)+(200<<1)+0 },
  { 0,"subfze.",F_SUPP_B,T_X,31,0,0,(0<<10)+(200<<1)+1 },
  { 0,"subfzeo",F_SUPP_B,T_X,31,0,0,(1<<10)+(200<<1)+0 },
  { 0,"subfzeo.",F_SUPP_B,T_X,31,0,0,(1<<10)+(200<<1)+1 },
  { 0,"sync",F_SUPP_D|F_SUPP_A|F_SUPP_B,T_X,31,0,0,(598<<1) },
  { 0,"td",F_64BIT,T_X,31,0,0,(68<<1) },
  { 0,"tdi",F_SIGNED|F_64BIT,T_DI,2,0,0,0 },
  { 0,"tlbia",F_SUPP_D|F_SUPP_A|F_SUPP_B|F_SUPER|F_OPTIONAL,
    T_X,31,0,0,(370<<1) },
  { 0,"tlbie",F_SUPP_D|F_SUPP_A|F_SUPER|F_OPTIONAL,T_X,31,0,0,(306<<1) },
  { 0,"tlbsync",F_SUPP_D|F_SUPP_A|F_SUPP_B|F_SUPER|F_OPTIONAL,
    T_X,31,0,0,(566<<1) },
  { 0,"tw",0,T_X,31,0,0,(4<<1) },
  { 0,"twi",F_SIGNED,T_DI,3,0,0,0 },
  { 0,"xor",F_SWAP,T_X,31,0,0,(316<<1)+0 },
  { 0,"xor.",F_SWAP,T_X,31,0,0,(316<<1)+1 },
  { 0,"xori",F_SWAP,T_DI,26,0,0,0 },
  { 0,"xoris",F_SWAP,T_DI,27,0,0,0 },

  /* extended mnemonics, rest is implemented as predefined macros */
  { 0,"la",F_SIGNED|F_EXTENDED,T_DD,14,0,0,0 },

  /* AltiVec mnemonics */
  { 0,"dss",F_ALTIVEC|F_SUPP_A|F_SUPP_B,T_DSTRM,31,0,0,(822<<1) },
  { 0,"dst",F_ALTIVEC,T_DSTRM,31,0,0,(342<<1) },
  { 0,"dstst",F_ALTIVEC,T_DSTRM,31,0,0,(374<<1) },
  { 0,"lvebx",F_ALTIVEC,T_X,31,0,0,(7<<1) },
  { 0,"lvehx",F_ALTIVEC,T_X,31,0,0,(39<<1) },
  { 0,"lvewx",F_ALTIVEC,T_X,31,0,0,(71<<1) },
  { 0,"lvsl",F_ALTIVEC,T_X,31,0,0,(6<<1) },
  { 0,"lvsr",F_ALTIVEC,T_X,31,0,0,(38<<1) },
  { 0,"lvx",F_ALTIVEC,T_X,31,0,0,(103<<1) },
  { 0,"lvxl",F_ALTIVEC,T_X,31,0,0,(359<<1) },
  { 0,"mfvscr",F_ALTIVEC|F_SUPP_A|F_SUPP_B,T_X,4,0,0,1540 },
  { 0,"mtvscr",F_ALTIVEC|F_SUPP_D|F_SUPP_A,T_X,4,0,0,1604 },
  { 0,"stvebx",F_ALTIVEC,T_X,31,0,0,(135<<1) },
  { 0,"stvehx",F_ALTIVEC,T_X,31,0,0,(167<<1) },
  { 0,"stvewx",F_ALTIVEC,T_X,31,0,0,(199<<1) },
  { 0,"stvx",F_ALTIVEC,T_X,31,0,0,(231<<1) },
  { 0,"stvxl",F_ALTIVEC,T_X,31,0,0,(487<<1) },
  { 0,"vaddcuw",F_ALTIVEC,T_VX,4,0,0,384 },
  { 0,"vaddfp",F_ALTIVEC,T_VX,4,0,0,10 },
  { 0,"vaddsbs",F_ALTIVEC,T_VX,4,0,0,768 },
  { 0,"vaddshs",F_ALTIVEC,T_VX,4,0,0,832 },
  { 0,"vaddsws",F_ALTIVEC,T_VX,4,0,0,896 },
  { 0,"vaddubm",F_ALTIVEC,T_VX,4,0,0,0 },
  { 0,"vaddubs",F_ALTIVEC,T_VX,4,0,0,512 },
  { 0,"vadduhm",F_ALTIVEC,T_VX,4,0,0,64 },
  { 0,"vadduhs",F_ALTIVEC,T_VX,4,0,0,576 },
  { 0,"vadduwm",F_ALTIVEC,T_VX,4,0,0,128 },
  { 0,"vadduws",F_ALTIVEC,T_VX,4,0,0,640 },
  { 0,"vand",F_ALTIVEC,T_VX,4,0,0,1028 },
  { 0,"vandc",F_ALTIVEC,T_VX,4,0,0,1092 },
  { 0,"vavgsb",F_ALTIVEC,T_VX,4,0,0,1282 },
  { 0,"vavgsh",F_ALTIVEC,T_VX,4,0,0,1346 },
  { 0,"vavgsw",F_ALTIVEC,T_VX,4,0,0,1410 },
  { 0,"vavgub",F_ALTIVEC,T_VX,4,0,0,1026 },
  { 0,"vavguh",F_ALTIVEC,T_VX,4,0,0,1090 },
  { 0,"vavguw",F_ALTIVEC,T_VX,4,0,0,1154 },
  { 0,"vcfsx",F_ALTIVEC,T_VX,4,0,0,842 },
  { 0,"vcfux",F_ALTIVEC,T_VX,4,0,0,778 },
  { 0,"vcmpbfp",F_ALTIVEC,T_VX,4,0,0,(0<<10)+966 },
  { 0,"vcmpeqfp",F_ALTIVEC,T_VX,4,0,0,(0<<10)+198 },
  { 0,"vcmpequb",F_ALTIVEC,T_VX,4,0,0,(0<<10)+6 },
  { 0,"vcmpeqhb",F_ALTIVEC,T_VX,4,0,0,(0<<10)+70 },
  { 0,"vcmpeqwb",F_ALTIVEC,T_VX,4,0,0,(0<<10)+134 },
  { 0,"vcmpgefp",F_ALTIVEC,T_VX,4,0,0,(0<<10)+454 },
  { 0,"vcmpgtfp",F_ALTIVEC,T_VX,4,0,0,(0<<10)+710 },
  { 0,"vcmpgtsb",F_ALTIVEC,T_VX,4,0,0,(0<<10)+774 },
  { 0,"vcmpgtsh",F_ALTIVEC,T_VX,4,0,0,(0<<10)+838 },
  { 0,"vcmpgtsw",F_ALTIVEC,T_VX,4,0,0,(0<<10)+902 },
  { 0,"vcmpgtub",F_ALTIVEC,T_VX,4,0,0,(0<<10)+518 },
  { 0,"vcmpgtuh",F_ALTIVEC,T_VX,4,0,0,(0<<10)+582 },
  { 0,"vcmpgtuw",F_ALTIVEC,T_VX,4,0,0,(0<<10)+646 },
  { 0,"vcmpbfp.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+966 },
  { 0,"vcmpeqfp.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+198 },
  { 0,"vcmpequb.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+6 },
  { 0,"vcmpeqhb.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+70 },
  { 0,"vcmpeqwb.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+134 },
  { 0,"vcmpgefp.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+454 },
  { 0,"vcmpgtfp.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+710 },
  { 0,"vcmpgtsb.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+774 },
  { 0,"vcmpgtsh.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+838 },
  { 0,"vcmpgtsw.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+902 },
  { 0,"vcmpgtub.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+518 },
  { 0,"vcmpgtuh.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+582 },
  { 0,"vcmpgtuw.",F_ALTIVEC,T_VX,4,0,0,(1<<10)+646 },
  { 0,"vctsxs",F_ALTIVEC,T_VX,4,0,0,842 },
  { 0,"vctuxs",F_ALTIVEC,T_VX,4,0,0,778 },
  { 0,"vexptefp",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,394 },
  { 0,"vlogefp",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,458 },
  { 0,"vmaddfp",F_ALTIVEC,T_VA,4,0,0,46 },
  { 0,"vmaxfp",F_ALTIVEC,T_VX,4,0,0,1034 },
  { 0,"vmaxsb",F_ALTIVEC,T_VX,4,0,0,258 },
  { 0,"vmaxsh",F_ALTIVEC,T_VX,4,0,0,322 },
  { 0,"vmaxsw",F_ALTIVEC,T_VX,4,0,0,386 },
  { 0,"vmaxub",F_ALTIVEC,T_VX,4,0,0,2 },
  { 0,"vmaxuh",F_ALTIVEC,T_VX,4,0,0,66 },
  { 0,"vmaxuw",F_ALTIVEC,T_VX,4,0,0,130 },
  { 0,"vmhaddshs",F_ALTIVEC,T_VA,4,0,0,32 },
  { 0,"vmhraddshs",F_ALTIVEC,T_VA,4,0,0,33 },
  { 0,"vminfp",F_ALTIVEC,T_VX,4,0,0,1098 },
  { 0,"vminsb",F_ALTIVEC,T_VX,4,0,0,770 },
  { 0,"vminsh",F_ALTIVEC,T_VX,4,0,0,834 },
  { 0,"vminsw",F_ALTIVEC,T_VX,4,0,0,898 },
  { 0,"vminub",F_ALTIVEC,T_VX,4,0,0,514 },
  { 0,"vminuh",F_ALTIVEC,T_VX,4,0,0,578 },
  { 0,"vminuw",F_ALTIVEC,T_VX,4,0,0,642 },
  { 0,"vmladduhm",F_ALTIVEC,T_VA,4,0,0,34 },
  { 0,"vmrghb",F_ALTIVEC,T_VX,4,0,0,12 },
  { 0,"vmrghh",F_ALTIVEC,T_VX,4,0,0,76 },
  { 0,"vmrghw",F_ALTIVEC,T_VX,4,0,0,140 },
  { 0,"vmrglb",F_ALTIVEC,T_VX,4,0,0,268 },
  { 0,"vmrglh",F_ALTIVEC,T_VX,4,0,0,332 },
  { 0,"vmrglw",F_ALTIVEC,T_VX,4,0,0,396 },
  { 0,"vmsummbm",F_ALTIVEC,T_VA,4,0,0,37 },
  { 0,"vmsumshm",F_ALTIVEC,T_VA,4,0,0,40 },
  { 0,"vmsumshs",F_ALTIVEC,T_VA,4,0,0,41 },
  { 0,"vmsumubm",F_ALTIVEC,T_VA,4,0,0,36 },
  { 0,"vmsumuhm",F_ALTIVEC,T_VA,4,0,0,38 },
  { 0,"vmsumuhs",F_ALTIVEC,T_VA,4,0,0,39 },
  { 0,"vmulesb",F_ALTIVEC,T_VX,4,0,0,776 },
  { 0,"vmulesh",F_ALTIVEC,T_VX,4,0,0,840 },
  { 0,"vmuleub",F_ALTIVEC,T_VX,4,0,0,520 },
  { 0,"vmuleuh",F_ALTIVEC,T_VX,4,0,0,584 },
  { 0,"vmulosb",F_ALTIVEC,T_VX,4,0,0,264 },
  { 0,"vmulosh",F_ALTIVEC,T_VX,4,0,0,328 },
  { 0,"vmuloub",F_ALTIVEC,T_VX,4,0,0,8 },
  { 0,"vmulouh",F_ALTIVEC,T_VX,4,0,0,72 },
  { 0,"vnmsubfp",F_ALTIVEC,T_VA,4,0,0,47 },
  { 0,"vnor",F_ALTIVEC,T_VX,4,0,0,1284 },
  { 0,"vor",F_ALTIVEC,T_VX,4,0,0,1156 },
  { 0,"vperm",F_ALTIVEC,T_VA,4,0,0,43 },
  { 0,"vpkpx",F_ALTIVEC,T_VX,4,0,0,782 },
  { 0,"vpkshss",F_ALTIVEC,T_VX,4,0,0,398 },
  { 0,"vpkshus",F_ALTIVEC,T_VX,4,0,0,270 },
  { 0,"vpkswss",F_ALTIVEC,T_VX,4,0,0,462 },
  { 0,"vpkswus",F_ALTIVEC,T_VX,4,0,0,334 },
  { 0,"vpkuhum",F_ALTIVEC,T_VX,4,0,0,14 },
  { 0,"vpkuhus",F_ALTIVEC,T_VX,4,0,0,142 },
  { 0,"vpkuwum",F_ALTIVEC,T_VX,4,0,0,78 },
  { 0,"vpkuwus",F_ALTIVEC,T_VX,4,0,0,206 },
  { 0,"vrefp",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,266 },
  { 0,"vrfim",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,714 },
  { 0,"vrfin",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,522 },
  { 0,"vrfip",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,650 },
  { 0,"vrfiz",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,586 },
  { 0,"vrlb",F_ALTIVEC,T_VX,4,0,0,4 },
  { 0,"vrlh",F_ALTIVEC,T_VX,4,0,0,68 },
  { 0,"vrlw",F_ALTIVEC,T_VX,4,0,0,132 },
  { 0,"vrsqrtefp",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,330 },
  { 0,"vsel",F_ALTIVEC,T_VA,4,0,0,42 },
  { 0,"vsl",F_ALTIVEC,T_VX,4,0,0,452 },
  { 0,"vslb",F_ALTIVEC,T_VX,4,0,0,260 },
  { 0,"vsldoi",F_ALTIVEC,T_VA,4,0,0,44 },
  { 0,"vslh",F_ALTIVEC,T_VX,4,0,0,324 },
  { 0,"vslo",F_ALTIVEC,T_VX,4,0,0,1036 },
  { 0,"vslw",F_ALTIVEC,T_VX,4,0,0,388 },
  { 0,"vspltb",F_ALTIVEC,T_VX,4,0,0,524 },
  { 0,"vsplth",F_ALTIVEC,T_VX,4,0,0,588 },
  { 0,"vspltw",F_ALTIVEC,T_VX,4,0,0,652 },
  { 0,"vspltisb",F_ALTIVEC|F_SUPP_B|F_SIGNED,T_VX,4,0,0,780 },
  { 0,"vspltish",F_ALTIVEC|F_SUPP_B|F_SIGNED,T_VX,4,0,0,844 },
  { 0,"vspltisw",F_ALTIVEC|F_SUPP_B|F_SIGNED,T_VX,4,0,0,908 },
  { 0,"vsr",F_ALTIVEC,T_VX,4,0,0,708 },
  { 0,"vsrab",F_ALTIVEC,T_VX,4,0,0,772 },
  { 0,"vsrah",F_ALTIVEC,T_VX,4,0,0,836 },
  { 0,"vsraw",F_ALTIVEC,T_VX,4,0,0,900 },
  { 0,"vsrb",F_ALTIVEC,T_VX,4,0,0,516 },
  { 0,"vsrh",F_ALTIVEC,T_VX,4,0,0,580 },
  { 0,"vsro",F_ALTIVEC,T_VX,4,0,0,1100 },
  { 0,"vsrw",F_ALTIVEC,T_VX,4,0,0,644 },
  { 0,"vsubcuw",F_ALTIVEC,T_VX,4,0,0,1408 },
  { 0,"vsubfp",F_ALTIVEC,T_VX,4,0,0,74 },
  { 0,"vsubsbs",F_ALTIVEC,T_VX,4,0,0,1792 },
  { 0,"vsubshs",F_ALTIVEC,T_VX,4,0,0,1856 },
  { 0,"vsubsws",F_ALTIVEC,T_VX,4,0,0,1920 },
  { 0,"vsububm",F_ALTIVEC,T_VX,4,0,0,1024 },
  { 0,"vsububs",F_ALTIVEC,T_VX,4,0,0,1536 },
  { 0,"vsubuhm",F_ALTIVEC,T_VX,4,0,0,1088 },
  { 0,"vsubuhs",F_ALTIVEC,T_VX,4,0,0,1600 },
  { 0,"vsubuwm",F_ALTIVEC,T_VX,4,0,0,1152 },
  { 0,"vsubuws",F_ALTIVEC,T_VX,4,0,0,1664 },
  { 0,"vsumsws",F_ALTIVEC,T_VX,4,0,0,1928 },
  { 0,"vsum2sws",F_ALTIVEC,T_VX,4,0,0,1672 },
  { 0,"vsum4sbs",F_ALTIVEC,T_VX,4,0,0,1800 },
  { 0,"vsum4shs",F_ALTIVEC,T_VX,4,0,0,1608 },
  { 0,"vsum4ubs",F_ALTIVEC,T_VX,4,0,0,1544 },
  { 0,"vupkhpx",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,846 },
  { 0,"vupkhsb",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,526 },
  { 0,"vupkhsh",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,590 },
  { 0,"vupklpx",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,974 },
  { 0,"vupklsb",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,654 },
  { 0,"vupklsh",F_ALTIVEC|F_SUPP_A,T_VX,4,0,0,718 },
  { 0,"vxor",F_ALTIVEC,T_VX,4,0,0,1220 },

  { 0,0 } /* end mark */
};



static void swapops(uint32 *a,uint32 *b)
{
  uint32 c = *a;

  *a = *b;
  *b = c;
}


static int count_operands(struct GlobalVars *gv,char *s)
{
  int n=0;

  if (*s!=0 && *s!='#')
    do {
      n++;
      s = skipexpression(gv,s);
    }
    while (*s++ == ',');
  return (n);
}


static bool chk_regindir(char *s)
/* "(...)\0" has a high probability to be a register-indirect */
/* addressing mode. */
{
  char c;

  if (*s++ == '(') {
    while (c = *s++) {
      if (c == '(')
        break;           /* "((..." will never be register-indirect */
      if (c == ')') {
        s = skipspaces(s);
        if (*s=='\0' || *s=='#')  /* line ends directly after "(...)" ? */
          return (TRUE);
        else
          break;
      }
    }
  }
  return (FALSE);
}


static void move_aligned_symbols_offset(struct GlobalVars *gv,
                                        struct AlignPoint *ap,
                                        uint32 so,int32 d)
{
  struct Symbol *sym;
  struct Symbol **symtab = gv->symbols;
  int i;

  for (i=0; i<SYMHTABSIZE; i++) {
    sym = *symtab++;
    while (sym) {
      if (sym->type == SYM_RELOC)
        if (sym->alignpoint == ap)
          if (sym->value > so)
            sym->value += d;
      sym = sym->hash_chain;
    }
  }
}


static void move_aligned_symbols_all(struct GlobalVars *gv,
                                     struct AlignPoint *ap,int32 d)
{
  struct Symbol *sym;
  struct Symbol **symtab = gv->symbols;
  int i;

  for (i=0; i<SYMHTABSIZE; i++) {
    sym = *symtab++;
    while (sym) {
      if (sym->type==SYM_RELOC && sym->alignpoint==ap)
        sym->value += d;
      sym = sym->hash_chain;
    }
  }
}


static void move_symbols(struct GlobalVars *gv,uint32 so,int32 d)
/* All relocatable symbols of current section with a greater */
/* section-offset than 'so' are moved by delta 'd'. */
{
  struct AlignPoint *ap = gv->csect->current_align;
  long g,a;
  unsigned long off;

  move_aligned_symbols_offset(gv,ap,so,d);
  while (ap = ap->next) {
    g = ap->gap - d;
    a = ap->val;
    off = ap->offset + ap->gap;  /* current offset after alignment */
    ap->offset += d;             /* new align-point offset */
    if (g<0 || g>=a) {  /* need to move the alignment point? */
      g = (a - (ap->offset & (a-1))) & (a-1);
      d = ap->offset + g - off;
      ap->gap = g;
      move_aligned_symbols_all(gv,ap,d);    
    }
    else {  /* delta is neutralized by alignment gap */
      ap->gap = g;
      break;
    }
  }
  gv->csect->data = NULL;  /* no further data is written to section */
  gv->anotherpass = TRUE;  /* this requires another pass :( */
}


static void opt_far_branch(struct GlobalVars *gv,struct ParsedLine *pl,
                           uint32 *bo,uint32 *bi,int32 *bd,char *lab)
/* If branch destination is out of range, convert into a */
/* B!cc $+8 / B combination. */
{
  if (*bd>0x7fff || *bd<-0x8000) {
    struct ParsedLine *newpl = alloczero(sizeof(struct ParsedLine));
    char buf[STRBUFSIZE];

    move_symbols(gv,(uint32)gv->csect->pc,4); /* space for new instruction */
    if (!(*bo & 16))
      *bo ^= 8;  /* negate CR condition */
    if (!(*bo & 4))
      *bo ^= 2;  /* negate CTR condition */
    *bd = 8;   /* branch displacement = "$+8" */
    newpl->type = OT_INSTRUCTION;
    newpl->flags = pl->flags;
    newpl->narg = pl->narg;
    newpl->lineptr = pl->lineptr;
    newpl->opcode = (void *)search_instr(gv,"b");
    newpl->operand = lab;
    newpl->next = pl->next;
    /* free(pl->operand); @@@@ ? */
    sprintf(buf,"%d,%d,$+8",(int)*bo,(int)*bi);
    pl->branch_hint = -(pl->branch_hint);
    pl->operand = allocstring(buf);
    pl->next = newpl;
  }
}


char *check_comma(char *s)
{
  s = skipspaces(s);
  if (*s++ != ',') {
    error(27);  /* instruction has too few operands */
    return (NULL);
  }
  s = skipspaces(s);
  return (s);
}


void pcadd(struct GlobalVars *gv,unsigned long offset)
{
  gv->csect->pc += offset;
}


void store_byte(struct GlobalVars *gv,uint8 x)
{
  struct Section *sec = gv->csect;
  uint8 *d;

  if (d = (uint8 *)sec->data) {
    *d++ = x;
    sec->data = d;
  }
  sec->pc += 1;
  gv->lcsym->value += 1;
}


void store_half(struct GlobalVars *gv,uint16 x)
{
  struct Section *sec = gv->csect;
  uint8 *d;

  if (d = (uint8 *)sec->data) {
    *d++ = (uint8)((x>>8)&0xff);
    *d++ = (uint8)(x&0xff);
    sec->data = d;
  }
  sec->pc += 2;
  gv->lcsym->value += 2;
}


void store_word(struct GlobalVars *gv,uint32 x)
{
  struct Section *sec = gv->csect;
  uint8 *d;

  if (d = (uint8 *)sec->data) {
    *d++ = (uint8)((x>>24)&0xff);
    *d++ = (uint8)((x>>16)&0xff);
    *d++ = (uint8)((x>>8)&0xff);
    *d++ = (uint8)(x&0xff);
    sec->data = d;
  }
  sec->pc += 4;
  gv->lcsym->value += 4;
}


void store_float(struct GlobalVars *gv,double x)
{
  struct Section *sec = gv->csect;
  float f = (float)x;
  uint8 *d,*p = (uint8 *)&f;

  if (d = (uint8 *)sec->data) {
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p;
    sec->data = d;
  }
  sec->pc += 4;
  gv->lcsym->value += 4;
}


void store_double(struct GlobalVars *gv,double x)
{
  struct Section *sec = gv->csect;
  uint8 *d,*p = (uint8 *)&x;

  if (d = (uint8 *)sec->data) {
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p++;
    *d++ = *p;
    sec->data = d;
  }
  sec->pc += 8;
  gv->lcsym->value += 8;
}


void store_space(struct GlobalVars *gv,unsigned long n)
{
  struct Section *sec = gv->csect;
  uint8 *d;

  sec->pc += n;
  gv->lcsym->value += n;
  if (d = (uint8 *)sec->data) {
    while (n--)
      *d++ = 0;
    sec->data = d;
  }
}


void instr(struct GlobalVars *gv,struct ParsedLine *pl)
{
  struct CPUInstr *ins=(struct CPUInstr *)pl->opcode;
  char *op=pl->operand;
  uint32 d=ins->fieldD,a=ins->fieldA;
  uint32 b=0,c=0,x=0;
  uint32 oc=ins->opcd<<26;
  uint32 xo=ins->xo;
  uint16 flags=ins->flags;
  char *oldop=op;
  struct Expression exp;

  if (flags & F_64BIT) {
    if (!gv->sixtyfourmode)
      error(37,ins->name);  /* warning: 64-bit instruction */
  }
  if (flags & F_OPTIONAL) {
    if (!gv->optinstrmode)
      error(38,ins->name);  /* warning: optional instruction */
  }
  if (flags & F_SUPER) {
    if (!gv->supermode)
      error(39,ins->name);  /* warning: supervisor-level instruction */
  }
  if (flags & F_ALTIVEC) {
    if (!gv->altivecmode)
      error(64,ins->name);  /* warning: AltiVec instruction */
  }
  gv->signedexp = flags&F_SIGNED ? TRUE:FALSE;

  switch (ins->type) {
  
    case T_I:  /* Bx */
      op = eval_expression(gv,&exp,op);
      if (exp.type < SYM_RELOC)
        switch (exp.type) {
          case SYM_UNDEF:
            error(19);  /* undefined symbol */
            break;
          case SYM_ABS:
            error(32);  /* Reloc symbol required */
            break;
        }
      if (flags & F_SIGNED) {  /* relative branch */
        exp.reloctype = R_PPC_REL24;
        switch (exp.type) {
          case SYM_RELOC:
            if (exp.symbol->relsect == gv->csect)
              x = exp.value - gv->csect->pc;  /* branch distance */
            else {
              error(33);  /* branch destination is not in current section */
              x = makereloc(gv,&exp);
            }
            break;
          case SYM_EXTERN:
          case SYM_COMMON:
            x = makexref(gv,&exp,3);
            break;
        }
        checkrange(x,3,TRUE);
      }
      else {  /* absolute branch */
        exp.reloctype = R_PPC_ADDR24;
        switch (exp.type) {
          case SYM_RELOC:
            x = makereloc(gv,&exp);
            break;
          case SYM_EXTERN:
          case SYM_COMMON:
            x = makexref(gv,&exp,3);
            break;
        }
        checkrange(x,3,FALSE);        
      }
      oc |= (x&0x3fffffc)|xo;
      break;

    case T_B:  /* BCx */
      op = getintexp(gv,op,&d);  /* BO, branch operand */
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&a);  /* BI, branch condition index */
      if (!(op = check_comma(op)))
        break;
      oldop = op;
      op = eval_expression(gv,&exp,op);
      if (exp.type < SYM_RELOC)
        switch (exp.type) {
          case SYM_UNDEF:
            error(19);  /* undefined symbol */
            break;
          case SYM_ABS:
            error(32);  /* Reloc symbol required */
            break;
        }
      if (flags & F_SIGNED) {  /* relative branch */
        exp.reloctype = R_PPC_REL14;
        switch (exp.type) {
          case SYM_RELOC:
            if (exp.symbol->relsect == gv->csect) {
              x = exp.value - gv->csect->pc;  /* branch distance */
              if (gv->opt & OPT_FAR_BRANCH)
                opt_far_branch(gv,pl,&d,&a,(int32 *)&x,oldop);
            }
            else {
              error(33);  /* branch destination is not in current section */
              x = makereloc(gv,&exp);
            }
            break;
          case SYM_EXTERN:
          case SYM_COMMON:
            x = makexref(gv,&exp,2);
            break;
        }
        checkrange(x,2,TRUE);
      }
      else {  /* absolute branch */
        exp.reloctype = R_PPC_ADDR14;
        switch (exp.type) {
          case SYM_RELOC:
            x = makereloc(gv,&exp);
            break;
          case SYM_EXTERN:
          case SYM_COMMON:
            x = makexref(gv,&exp,2);
            break;
        }
        checkrange(x,2,FALSE);        
      }
      if (pl->branch_hint > 0) { /* + */
        if (flags & F_SIGNED) {
          if ((int32)x >= 0)
            d |= 1;
          else
            d &= ~1;
        }
        else {
          if (x >= gv->csect->pc && exp.symbol->relsect == gv->csect)
            d |= 1;
          else
            d &= ~1;
        }
      }
      else if (pl->branch_hint < 0) { /* - */
        if (flags & F_SIGNED) {
          if ((int32)x < 0)
            d |= 1;
          else
            d &= ~1;
        }
        else {
          if (x < gv->csect->pc && exp.symbol->relsect == gv->csect)
            d |= 1;
          else
            d &= ~1;
        }
      }
      x &= 0xfffc;
      if (gv->output == OFMT_EHF) {  /* hack to allow correct EHF reloc. */
        if ((flags&F_SIGNED) && (exp.type==SYM_EXTERN ||
            (exp.type==SYM_RELOC && exp.symbol->relsect!=gv->csect)))
          x += 2;
      }
      if (d>31 || a>31)
        error(34);  /* Branch operand out of range */
      else
        oc |= (d<<21)|(a<<16)|x|xo;
      break;

    case T_XLB:  /* BCLRx,BCCTRx */
      op = getintexp(gv,op,&d);  /* BO, branch operand */
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&a);  /* BI, branch condition index */
      if (pl->branch_hint > 0)
        d |= 1;
      else if (pl->branch_hint < 0)
        d &= ~1;
      if (d>31 || a>31)
        error(34);  /* Branch operand out of range */
      else
        oc |= (d<<21)|(a<<16)|xo;
      break;

    case T_CMP:  /* CMP, CMPI, CMPL, CMPLI */
      if (count_operands(gv,op) > 3) {  /* CR-field specified? */
        op = getintexp(gv,op,&d);
        if (!(op = check_comma(op)))
          break;
      }
      op = getintexp(gv,op,&c);  /* L = word or double word */
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&a);
      if (!(op = check_comma(op)))
        break;
      if (flags & F_SUPP_B) {  /* immediate comparison */
        op = getexp(gv,op,&x,2);
        checkrange(x,2,flags&F_SIGNED);
      }
      else  /* register comparison */
        op = getintexp(gv,op,&b);
      if (d>7)
        error(35);  /* CR-specifier out of range */
      else if (c>1)
        error(36,"L");  /* Operand field out of range */
      else if (a>31 || b>31)
        error(31);  /* register operand out of range */
      else if (flags & F_SUPP_B)
        oc |= (d<<23)|(c<<21)|(a<<16)|(x&0xffff);
      else
        oc |= (d<<23)|(c<<21)|(a<<16)|(b<<11)|xo;
      break;

    case T_DI:  /* ADDI,ANDI,ORI */
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&a);
      if (!(op = check_comma(op)))
        break;
      op = getexp(gv,op,&x,2);
      if (flags & F_SWAP)
        swapops(&d,&a);
      checkrange(x,2,flags&F_SIGNED);
      if (d>31 || a>31)
        error(31);  /* register operand out of range */
      else
        oc |= (d<<21)|(a<<16)|(x&0xffff);
      break;

    case T_X:  /* ADDx, ANDx, DIVx, FMRx, .. */
    case T_VX: /* VADDxxx, VANDx, VAVGxx, VMAXxx, VMINxx, VMULxx, VSUBxxx */
      if (!(flags & F_SUPP_D))
        op = getintexp(gv,op,&d);
      if (!(flags & F_SUPP_A)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&a);
      }
      if (!(flags & F_SUPP_B)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&b);
      }
      if (flags & F_CRF_D)
        d <<= 2;
      if (flags & F_CRF_S)
        a <<= 2;
      if ((d==a || d==b) && (flags & F_UPDATE))
        error(58,d);  /* warning: rD = rA */
      if (flags & F_SWAP)
        swapops(&d,&a);
      if ((ins->type==T_VX && (flags & F_SIGNED)) &&
          (d>31 || a<-16 || a>15 || b>31))
        error(31);
      else if (d>31 || a>31 || b>31)
        error(31);
      else
        oc |= (d<<21)|(a<<16)|(b<<11)|xo;
      break;

    case T_A:  /* FADDx, FSUBx, FMULx, FDIVx, .. */
    case T_VA: /* VMADDFP, VMHADDSHS, VMSUMMBM, .. */
      if (!(flags & F_SUPP_D))
        op = getintexp(gv,op,&d);
      if (!(flags & F_SUPP_A)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&a);
      }
      if (!(flags & F_SUPP_C)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&c);
      }
      if (!(flags & F_SUPP_B)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&b);
      }
      if (flags & F_SWAP) {  /* rldcl/rldcr rA,rS,rB,mb */
        swapops(&d,&a);
        swapops(&c,&b);
      }
      if (ins->type == T_VA)
        swapops(&c,&b);   /* T_VA has always the form: vD,vA,vB,vC */
      if (ins->type==T_VA && xo==44 && c>15)  /* special case: VSLDOI */
        error(36,"SH");  /* operand field out of range */
      else if (d>31 || a>31 || b>31 || c>31)
        error(31);
      else
        oc |= (d<<21)|(a<<16)|(b<<11)|(c<<6)|xo;
      break;

    case T_DD:  /* LBZ, LHA, LZWU, STBU, STW, .. */
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      if (chk_regindir(op)) {   /* rD,(rA) -> rD,0(rA) */
        exp.value = 0;
        exp.type = SYM_ABS;
      }
      else
        op = eval_expression(gv,&exp,op);
      if (*op++ == '(') {
        op = getintexp(gv,op,&a);
        if (*op++ != ')') {
          error(40);  /* (rA) expected, after index expression */
          op = NULL;
          break;
        }
        if (exp.type != SYM_UNDEF) {
          switch (exp.type) {
            case SYM_RELOC:
#if 0 /* @@@ TEST - Is this really needed??? @@@ */
              if (exp.reloctype==R_PPC_ADDR16_LO ||
                  exp.reloctype==R_PPC_SDAREL16) {  /* sym@l, sym@sdarx */
                x = makereloc(gv,&exp);
              }
              else {
                if ((int)a != gv->rtoc) {  /* valid ref. to rtoc symbol? */
                  /* Wrong base reg. for base rel. reloc sym. access */
                  error(46);
                }
                exp.reloctype = R_PPC_SDAREL16;
                x = makereloc(gv,&exp);
              }
#else
              if (elf_relocsize(exp.reloctype) > 2) {
                if (exp.reloctype==R_PPC_ADDR32 && gv->rtoc==(int)a)
                  /* small data symbol via small data base register */
                  exp.reloctype = R_PPC_SDAREL16;
                else
                  /* Wrong base reg. for base rel. reloc sym. access */
                  error(46);
              }
              x = makereloc(gv,&exp);
#endif
              break;
            case SYM_EXTERN:  /* reference to external symbol */
            case SYM_COMMON:
#if 0 /* @@@ TEST - Is this really needed??? @@@ */
              if (exp.reloctype!=R_PPC_ADDR16_LO &&
                  exp.reloctype!=R_PPC_SDAREL16) {
                if (gv->rtoc == (int)a)            /* ext. symbol via rtoc -  */
                  exp.reloctype = R_PPC_SDAREL16;  /* might be a SDAREL16 ref.*/
                else {
                  exp.reloctype = R_PPC_ADDR16;
                  gv->csect->pc += 2;
                  x = makexref(gv,&exp,2);
                  gv->csect->pc -= 2;
                  break;
                }
              }
#else
              if (elf_relocsize(exp.reloctype) > 2) {
                if (gv->rtoc==(int)a) {
                  /* might be a small data symbol */
                  exp.reloctype = R_PPC_SDAREL16;
                }
                else {
                  /* limit external symbol to 16 bit relocation */
                  exp.reloctype = R_PPC_ADDR16;
                  gv->csect->pc += 2;
                  x = makexref(gv,&exp,2);
                  gv->csect->pc -= 2;
                  break;
                }
              }
#endif
              x = makexref(gv,&exp,2);
              break;
            default:  /* SYM_ABS */
              x = exp.value;
              break;
          }
        }
        else
          error(19);  /* undefined symbol */
      }
      else {  /* rD,v -> rD,v_off(rtoc) */
        op--;
        if (exp.type != SYM_UNDEF) {
          switch (exp.type) {
            case SYM_RELOC:
              a = gv->rtoc;
              exp.reloctype = R_PPC_SDAREL16;
              x = makereloc(gv,&exp);
              break;
            case SYM_EXTERN:  /* we assume that this is a SDAREL16 reference */
            case SYM_COMMON:
              a = gv->rtoc;
              exp.reloctype = R_PPC_SDAREL16;
              x = makexref(gv,&exp,2);
              break;
            default:
              error(32);  /* reloc symbol required */
              break;
          }
        }
        else
          error(19);
      }
      checkrange(x,2,TRUE);
      if (d == a && (flags & F_UPDATE))
        error(58,d);  /* warning: rD = rA */
      if (d>31 || a>31)
        error(31);  /* register operand out of range */
      else
        oc |= (d<<21)|(a<<16)|(x&0xffff);
      break;

    case T_DS:  /* LD, LDU, LWA, STD , .. */
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      if (chk_regindir(op)) {   /* rD,(rA) -> rD,0(rA) */
        exp.value = 0;
        exp.type = SYM_ABS;
      }
      else
        op = eval_expression(gv,&exp,op);
      if (*op++ == '(') {
        op = getintexp(gv,op,&a);
        if (*op++ != ')') {
          error(40);  /* (rA) expected, after index expression */
          op = NULL;
          break;
        }
        if (exp.type != SYM_UNDEF) {
          switch (exp.type) {
            case SYM_RELOC:
#if 0 /* @@@ TEST - Is this really needed??? @@@ */
              if (exp.reloctype==R_PPC_ADDR16_LO ||
                  exp.reloctype==R_PPC_SDAREL16) {  /* sym@l, sym@sdarx */
                x = makereloc(gv,&exp);
              }
              else {
                if ((int)a != gv->rtoc) {  /* valid ref. to rtoc symbol? */
                  /* Wrong base reg. for base rel. reloc sym. access */
                  error(46);
                }
                exp.reloctype = R_PPC_SDAREL16;
                x = makereloc(gv,&exp);
              }
#else
              x = makereloc(gv,&exp); /* @@@ TEST @@@ */
#endif
              break;
            case SYM_EXTERN:  /* reference to external symbol */
            case SYM_COMMON:
#if 0 /* @@@ TEST - Is this really needed??? @@@ */
              if (exp.reloctype!=R_PPC_ADDR16_LO &&
                  exp.reloctype!=R_PPC_SDAREL16) {
                if (gv->rtoc == (int)a)            /* ext. symbol via rtoc -  */
                  exp.reloctype = R_PPC_SDAREL16;  /* might be a SDAREL16 ref.*/
                else {
                  exp.reloctype = R_PPC_ADDR16;
                  gv->csect->pc += 2;
                  x = makexref(gv,&exp,2);
                  gv->csect->pc -= 2;
                  break;
                }
              }
#endif
              x = makexref(gv,&exp,2);
              break;
            default:  /* SYM_ABS */
              x = exp.value;
              break;
          }
        }
        else
          error(19);  /* undefined symbol */
      }
      else {  /* rD,v -> rD,v_off(rtoc) */
        op--;
        if (exp.type != SYM_UNDEF) {
          switch (exp.type) {
            case SYM_RELOC:
              a = gv->rtoc;
              exp.reloctype = R_PPC_SDAREL16;
              x = makereloc(gv,&exp);
              break;
            case SYM_EXTERN:  /* we assume that this is a SDAREL16 reference */
            case SYM_COMMON:
              a = gv->rtoc;
              exp.reloctype = R_PPC_SDAREL16;
              x = makexref(gv,&exp,2);
              break;
            default:
              error(32);  /* reloc symbol required */
              break;
          }
        }
        else
          error(19);
      }
      checkrange(x,2,TRUE);
      if (x & 3)
        error(41);  /* expression must be dividable by four */
      else if (d>31 || a>31)
        error(31);  /* register operand out of range */
      else
        oc |= (d<<21)|(a<<16)|(x&0xfffc)|xo;
      break;

    case T_XSPR:  /* MFSPR, MTSPR, MFTB */
      op = getintexp(gv,op,&d);
      if (count_operands(gv,op)==0 && xo==(371<<1)) {  /* mftb */
        /* mftb simplified mnemonic allows omitting the reg. number */
        a = 268;
      }
      else {
        if (!(op = check_comma(op)))
          break;
        op = getintexp(gv,op,&a);
      }
      if (flags & F_SWAP)
        swapops(&d,&a);
      if (d>31 || a>0x3ff)
        error(31);
      else
        oc |= (d<<21)|((a&0x1f)<<16)|((a&0x3e0)<<6)|xo;
      break;

    case T_XCRM:  /* MFCRF */
      op = getintexp(gv,op,&x);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&d);
      if (x > 0xff)
        error(42);  /* invalid register field mask */
      else if (a>31)
        error(31);
      else
        oc |= (d<<21)|(x<<12)|xo;
      break;

    case T_XFL:  /* MTFSF */
      op = getintexp(gv,op,&x);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&b);
      if (x > 0xff)
        error(42);  /* invalid register field mask */
      else if (b>31)
        error(31);
      else
        oc |= (x<<17)|(b<<11)|xo;
      break;

    case T_IMM:  /* MTFSFI */
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&b);
      if (b>15)
        error(36,"IMM");  /* operand field out of range */
      else if (d>7)
        error(31);
      else
        oc |= (d<<23)|(b<<12)|xo;
      break;

    case T_MD:  /* RLDIC, .. */
      op = getintexp(gv,op,&a);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&b);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&c);
      if (b>63)
        error(36,"SH");  /* operand field out of range */
      else if (c>63)
        error(36,"MB/ME");
      else if (d>31 || a>31)
        error(31);
      else
        oc |= (d<<21)|(a<<16)|((b&0x1f)<<11)|(c<<5)|xo|((b&0x20)>>4);
      break;

    case T_M:  /* RLWINM, RLWNM, RLWIMI, .. */
      op = getintexp(gv,op,&a);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&b);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&c);
      op = skipspaces(op);
      if (*op==',') {
        op = skipspaces(++op);
        op = getintexp(gv,op,&x);
      }
      else {
        uint32 mask,val=c;
        int i;

        for (i=0,mask=1; i<32; i++,mask<<=1) {
          if (val & mask) {
            x = 31 - i;
            break;
          }
        }
        for (; i<32; i++,mask<<=1) {
          if (!(val & mask))
            break;
        }
        c = 32 - i;
      }
      if (b>31)
        error(36,"SH/rB");  /* operand field out of range */
      else if (c>31)
        error(36,"MB");
      else if (x>31)
        error(36,"ME");
      else if (d>31 || a>31)
        error(31);
      else
        oc |= (d<<21)|(a<<16)|(b<<11)|(c<<6)|(x<<1)|xo;
      break;

    case T_XS:  /* SRADIx */
      op = getintexp(gv,op,&a);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&d);
      if (!(op = check_comma(op)))
        break;
      op = getintexp(gv,op,&b);
      if (b>63)
        error(36,"SH");  /* operand field out of range */
      else if (d>31 || a>31)
        error(31);
      else
        oc |= (d<<21)|(a<<16)|((b&0x1f)<<11)|xo|((b&0x20)>>4);
      break;

    case T_DSTRM:  /* DSS, DST, DSTST */
      if (!(flags & F_SUPP_A))
        op = getintexp(gv,op,&a);
      if (!(flags & F_SUPP_B)) {
        if (op != oldop)
          if (!(op = check_comma(op)))
            break;
        op = getintexp(gv,op,&b);
      }
      if (op != oldop)
        if (!(op = check_comma(op)))
          break;
      op = getintexp(gv,op,&d);
      op = skipspaces(op);
      if (*op!=0 && *op!='#') {
        /* evaluate T-field, which might be ommitted for ext. mnemonics */
        if (!(op = check_comma(op)))
          break;
        op = getintexp(gv,op,&x);
      }
      if (d>3)
        error(36,"STRM");  /* operand field out of range */
      else if (x>1)
        error(36,"T");
      else if (a>31 || b>31)
        error(31);  /* register operand out of range */
      else
        oc |= (x<<25)|(d<<21)|(a<<16)|(b<<11)|xo;
      break;
  }

  store_word(gv,oc);
  if (op)
    checkEOL(op);
}
